"""Main module for the project."""

import time
from argparse import Namespace
from contextlib import suppress
from pathlib import Path
from typing import Callable

from rich import print

from pyfindr.cli import exit_session, get_parsed_args, print_parser_help
from pyfindr.consts import EXIT_FAILURE, EXIT_SUCCESS
from pyfindr.search import (
    print_match_in_file,
    print_match_in_filename,
    rec_find,
    search_in_file,
    search_in_filename,
)


def main() -> None:
    """Main function"""
    args: Namespace = get_parsed_args()

    if args.mode == "contents":
        print("\n[green]Searching contents...[/]")
        search_fun: Callable[[Path, str], tuple[str, bool]] = search_in_file
        print_fun: Callable[[str, str], None] = print_match_in_file

    elif args.mode == "filenames":
        print("\n[green]Searching filenames...[/]")
        search_fun = search_in_filename
        print_fun = print_match_in_filename

    else:
        print_parser_help()
        exit_session(EXIT_SUCCESS)

    search_dir: Path = Path(args.path).resolve()

    print()

    try:
        start_time: float = time.time()
        for fname in search_dir.iterdir():
            with suppress(Exception):
                rec_find(
                    path=fname,
                    key=args.key,
                    max_depth=args.max_depth,
                    search_fun=search_fun,
                    print_fun=print_fun,
                    no_dotfiles=args.skip_dotfiles,
                )

        if args.mode == "filenames":
            print()

        end_time: float = time.time()
        print(f"[green]Search completed in {end_time - start_time:.5f} seconds[/]")

        exit_session(EXIT_SUCCESS)

    except KeyboardInterrupt:
        print("\n[red]Cancelled[/]")

        exit_session(EXIT_FAILURE)


if __name__ == "__main__":
    main()



"""Command-line interface for the project."""

from pathlib import Path
import sys
from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from typing import NoReturn

from rich import print

from pyfindr.consts import EXIT_FAILURE, PACKAGE
from pyfindr.consts import __desc__ as DESC
from pyfindr.consts import __version__ as VERSION


parser: ArgumentParser


def get_parsed_args() -> Namespace:
    """
    Parse and return command-line arguments.

    Returns:
        argparse.Namespace: The parsed arguments as a Namespace object.
    """
    global parser

    parser = ArgumentParser(
        description=DESC,  # Program description
        formatter_class=RawTextHelpFormatter,  # Custom formatter
        allow_abbrev=False,  # Disable abbreviations
        add_help=False,  # Disable default help
    )

    g_main = parser.add_argument_group("Main Options")

    # key argument
    g_main.add_argument(
        "key",
        type=str,
        help="The string to search for.",
    )
    # path argument
    default_path: Path = Path.cwd()
    g_main.add_argument(
        "--path",
        default=default_path,
        help=f"the path to search under (default: {default_path})",
    )
    # mode argument
    default_mode: str = "contents"
    g_main.add_argument(
        "--mode",
        type=str,
        choices=["contents", "filenames"],
        default=default_mode,
        help=f"The search mode. Default is '{default_mode}'.",
    )
    # max-depth argument
    g_main.add_argument(
        "--max-depth",
        type=int,
        default=999,
        help="maximum depth for recursive search",
    )
    # skip-dotfiles argument
    g_main.add_argument(
        "--skip-dotfiles",
        action="store_true",
        default=False,
        help="skip dotfiles",
    )

    g_misc = parser.add_argument_group("Miscellaneous Options")
    # Help
    g_misc.add_argument(
        "-h", "--help", action="help", help="Show this help message and exit."
    )
    # Verbose
    g_misc.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="Show log messages on screen. Default is False.",
    )
    # Debug
    g_misc.add_argument(
        "-d",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Activate debug logs. Default is False.",
    )
    g_misc.add_argument(
        "-V",
        "--version",
        action="version",
        help="Show version number and exit.",
        version=f"[argparse.prog]{PACKAGE}[/] version [i]{VERSION}[/]",
    )

    return parser.parse_args()


def print_parser_help() -> None:
    """Print the help message for the parser."""
    global parser

    parser.print_help()


def exit_session(exit_value: int) -> NoReturn:
    """
    Exit the program with the given exit value.

    Args:
        exit_value (int): The POSIX exit value to exit with.

    Returns:
        NoReturn: This function does not return anything
    """
    # Check if the exit_value is a valid POSIX exit value
    if not 0 <= exit_value <= 255:
        exit_value = EXIT_FAILURE

    if exit_value == EXIT_FAILURE:
        print(
            "\n[red]There were errors during the execution of the script.[/]",
        )

    # Exit the program with the given exit value
    sys.exit(exit_value)



"""Constants for the project."""

try:
    from importlib import metadata
except ImportError:  # for Python < 3.8
    import importlib_metadata as metadata  # type: ignore

__version__: str = metadata.version(__package__ or __name__)
__desc__: str = metadata.metadata(__package__ or __name__)["Summary"]
PACKAGE: str = metadata.metadata(__package__ or __name__)["Name"]
GITHUB: str = metadata.metadata(__package__ or __name__)["Home-page"]

HIGHLIGHT_MAX_LEN = 40

EXIT_SUCCESS = 0
EXIT_FAILURE = 1

DEBUG = False
PROFILE = False



"""Search functions for the project."""

from contextlib import suppress
from pathlib import Path
from typing import Callable

from rich import print

from pyfindr.consts import HIGHLIGHT_MAX_LEN


def readlines(file: str | Path) -> list[str]:
    """
    Reads a file and returns its contents as a list of lines.

    Args:
        file (str | Path): The name of the file to read.

    Returns:
        list[str]: The contents of the file.
    """
    with open(file, "r", encoding="utf-8") as f:
        return f.readlines()


def print_match_in_file(fname: str, buffer: str) -> None:
    """
    Used to print a match found in a file.

    Args:
        fname (str): Name of file where match was found
        buffer (str): Buffer to print containing the match itself
    """
    print(f"[yellow]{fname}[/]\n{buffer}", flush=True)


def print_match_in_filename(_: str, buffer: str) -> None:
    """
    Used to print a match found in a filename.

    Args:
        _ (str): Name of file where match was found (not used)
        buffer (str): Buffer to print containing the match itself
    """
    print(buffer, flush=True)


def rec_find(  # pylint: disable=too-many-arguments
    path: Path,
    key: str,
    max_depth: int = 999,
    *,
    search_fun: Callable[[Path, str], tuple[str, bool]],
    print_fun: Callable[[str, str], None],
    no_dotfiles: bool = False,
) -> None:
    """
    Recursively search for key in file or directory.
    Calls search_fun for each file or directory found and print_fun
    for each match found.

    Args:
        path (Path): File or directory name
        key (str): Key to search
        max_depth (int, optional): Maximum depth to search (default: 999)
        search_fun (Callable[[Path, str], tuple[str, bool]]): Function to search for key
        print_fun (Callable[[str, str], None]): Function to print results
        no_dotfiles (bool, optional): Skip dotfiles (default: False)
    """
    if max_depth <= 0:
        return

    if no_dotfiles and path.name.startswith("."):
        return

    if path.is_file():
        with suppress(Exception):
            buffer, found = search_fun(path, key)
            if found:
                # Get the relative path
                relative_path = path.relative_to(Path.cwd())
                print_fun(str(relative_path), buffer)

    elif path.is_dir():
        for file in path.iterdir():
            rec_find(
                path=file,
                key=key,
                max_depth=max_depth - 1,
                search_fun=search_fun,
                print_fun=print_fun,
                no_dotfiles=no_dotfiles,
            )


def search_in_filename(path: Path, key: str) -> tuple[str, bool]:
    """
    Search for key in filename and return results

    Args:
        path (Path): File or directory name
        key (str): Key to search

    Returns:
        tuple[str, bool]: Buffer to store results, True if key is found
    """
    fname: str = path.name
    if key not in fname:
        return "", False

    start_idx = fname.index(key)
    end_idx = start_idx + len(key)
    buffer = (
        f"{fname[:start_idx]}[green]{fname[start_idx:end_idx]}" f"[/]{fname[end_idx:]}"
    )

    return buffer, True


def search_in_file(path: Path, key: str) -> tuple[str, bool]:
    """
    Search for key in file and store results in buffer

    Args:
        path (Path): File or directory name
        key (str): Key to search

    Returns:
        tuple[str, bool]: Buffer to store results, True if key is found
    """
    buffer = []
    for line_num, line in enumerate(readlines(path)):

        if key not in line:
            continue

        key_text = f"[green]{key}[/]"
        higlight = line.strip().replace(key, key_text)

        if higlight.index(key) > 20:
            higlight = higlight[higlight.index(key) - 20 :]

        location_text = (
            f"[blue]Line {line_num + 1}, "
            f"Column {line.find(key) + 1}:[/] "
            f"{higlight[:HIGHLIGHT_MAX_LEN]}"
        )

        if location_text[-1] != "\n":
            location_text = f"{location_text}\n"

        buffer.append(location_text)

    return "".join(buffer), len(buffer) > 0
